@@ -279,6 +279,9 @@ Entry to this mode runs the hooks on comint-mode-hook"
 ; (define-key comint-mode-map "\eP" 'comint-msearch-input)
 ; (define-key comint-mode-map "\eN" 'comint-psearch-input)
 ; (define-key comint-mode-map "\C-cR" 'comint-msearch-input-matching)
+  (define-key comint-mode-map "\C-c\C-n" 'comint-next-prompt)
+  (define-key comint-mode-map "\C-c\C-p" 'comint-prev-prompt)
+  (define-key comint-mode-map "\C-c\C-d" 'comint-send-eof)
   )
 
 
@@ -410,7 +413,7 @@ buffer.  The hook `comint-exec-hook' is run after each exec."
 		     (if (> arg 0) -1
 			 (if (< arg 0) 1 0))))
 	   (setq comint-input-ring-index
-		 (ring-mod (+ comint-input-ring-index arg) len))
+		 (mod (+ comint-input-ring-index arg) len))
 	   (message "%d" (1+ comint-input-ring-index))
 	   (insert (ring-ref comint-input-ring comint-input-ring-index))))))
 
@@ -447,11 +450,11 @@ If N is negative, find the next or Nth next match."
     (while (/= arg 0)
       (let ((prev n))
 	;; Step once.
-	(setq n (ring-mod (+ n motion) len))
+	(setq n (mod (+ n motion) len))
 	;; If we haven't reached a match, step some more.
 	(while (and (< n len)
 		    (not (string-match regexp (ring-ref comint-input-ring n))))
-	  (setq n (ring-mod (+ n motion) len))
+	  (setq n (mod (+ n motion) len))
 	  ;; If we have gone all the way around in this search, error.
 	  (if (= n prev)
 	      (error "Not found"))))
@@ -899,8 +902,29 @@ Useful if you accidentally suspend the top-level process."
       (process-send-eof)
       (delete-char arg)))
 
+(defun comint-send-eof ()
+  "Send an EOF to the current buffer's process."
+  (interactive)
+  (process-send-eof))
 
+(defun comint-next-prompt (n)
+  "\
+Move to end of next prompt in the buffer (with prefix arg, Nth next).
+See `comint-prompt-regexp'."
+  (interactive "p")
+  (re-search-forward comint-prompt-regexp nil nil n))
 
+(defun comint-prev-prompt (n)
+  "\
+Move to end of previous prompt in the buffer (with prefix arg, Nth previous).
+See `comint-prompt-regexp'."
+  (interactive "p")
+  (if (= (save-excursion (re-search-backward comint-prompt-regexp nil t)
+			 (match-end 0))
+	 (point))
+      (setq n (1+ n)))
+  (re-search-backward comint-prompt-regexp nil nil n)
+  (goto-char (match-end 0)))
 
 ;;; Support for source-file processing commands.
 ;;;============================================================================
@@ -1144,7 +1168,7 @@ See functions `expand-file-name' and `substitute-in-file-name'.  See also
 	   (message "No completions of %s" pathname)
 	   (ding))
 	  ((eql completion t)
-	   (message "Unique completion"))
+	   (message "Sole completion"))
 	  (t				; this means a string was returned.
 	   (delete-region (match-beginning 0) (match-end 0))
 	   (insert (expand-file-name (concat pathdir completion)))))))
@@ -1156,19 +1180,30 @@ This function is similar to `comint-replace-by-expanded-filename', except
 that it won't change parts of the filename already entered in the buffer; 
 it just adds completion characters to the end of the filename."
   (interactive)
-  (let* ((pathname (comint-match-partial-pathname))
-	 (pathdir (file-name-directory pathname))
-	 (pathnondir (file-name-nondirectory pathname))
-	 (completion (file-name-completion  pathnondir
-					   (or pathdir default-directory))))
-    (cond ((null completion)
-	   (message "No completions of %s" pathname)
-	   (ding))
-	  ((eql completion t)
-	   (message "Unique completion"))
-	  (t				; this means a string was returned.
-	   (goto-char (match-end 0))
-	   (insert (substring completion (length pathnondir)))))))
+  (if (and (interactive-p)
+	   (eq last-command this-command))
+      ;; If you hit TAB twice in a row, you get a completion list.
+      (comint-dynamic-list-completions)
+    (let* ((pathname (comint-match-partial-pathname))
+	   (pathdir (file-name-directory pathname))
+	   (pathnondir (file-name-nondirectory pathname))
+	   (completion (file-name-completion
+			pathnondir
+			;; It is important to expand PATHDIR because
+			;; default-directory might be a handled name, and the
+			;; unexpanded PATHDIR won't necessarily match the
+			;; handler regexp.
+			(if pathdir
+			    (expand-file-name pathdir)
+			  default-directory))))
+      (cond ((null completion)
+	     (message "No completions of %s" pathname)
+	     (ding))
+	    ((eql completion t)
+	     (message "Sole completion"))
+	    (t				; this means a string was returned.
+	     (goto-char (match-end 0))
+	     (insert (substring completion (length pathnondir))))))))
 
 (defun comint-dynamic-list-completions ()
   "List in help buffer all possible completions of the filename at point."
@@ -1178,7 +1213,9 @@ it just adds completion characters to the end of the filename."
 	 (pathnondir (file-name-nondirectory pathname))
 	 (completions
 	  (file-name-all-completions pathnondir
-				     (or pathdir default-directory))))
+				     (if pathdir
+					 (expand-file-name pathdir)
+				       default-directory))))
     (cond ((null completions)
 	   (message "No completions of %s" pathname)
 	   (ding))
